/***************************************************************************
                          dbgbase_classes.h  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
    email                : dd@cron.ru
    www                  : http://dd.cron.ru
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifndef _DBGBASE_CLASSES_H_
#define _DBGBASE_CLASSES_H_

//#include <string.h>
#include <string>
#include <vector>
#include <list>
#include <vector>

#ifdef WIN32
#else
#include <signal.h>
#endif

#include <dbgbase/dbg_net.h>
#include <dbgbase/dbgbase_globals.h>


/**************************************************************************

	RequestorBase

**************************************************************************/

class RequestorBase {
private:
protected:
public:
	virtual ~RequestorBase() {}
/* info */
	virtual HRESULT getdbgversion(
			/* [out] */ long *dbg_ver,
			/* [out] */ std::string *sdbg_descr) = 0;
	virtual HRESULT getserveraddr(
			/* [out] */ sockaddr_in *addr) = 0;
	virtual HRESULT setsessionoptions(
			/* [in] */ long opt_flags) = 0;
/* modules */
    virtual HRESULT requestdocsrc(
            /* [in] */  long mod_no,
            /* [out] */ std::string *mod_name,
            /* [out] */ std::string *src) = 0;
    virtual HRESULT requestdocparent( 
            /* [in] */ long mod_no,
            /* [out] */ long *parent_mod_no) = 0;
    virtual HRESULT requestdocfilename( 
            /* [in] */ long mod_no,
            /* [out] */ std::string *mod_name) = 0;
	virtual HRESULT getdoclist(
			/* [in/out] */ DOCLIST *doclist) = 0;
	virtual HRESULT getlinesinfolist(
            /* [in] */ long mod_no,
			/* [in/out] */ LINESINFOLIST *linesinfolist) = 0;
	virtual HRESULT getctxinfolist(
            /* [in] */ long mod_no,
			/* [in/out] */ CTXINFOLIST *ctxinfolist) = 0;
/* scopes */
	virtual HRESULT getscopelist(
			/* [in/out] */ SCOPELIST *ascopelist) = 0; 
	virtual HRESULT getcurloc(
			/* [out] */ long *mod_no,
			/* [out] */ long *line_no,
			/* [out] */ std::string *scope_name) = 0;
/* breakpoints */
	virtual HRESULT setbreakpoint(
			/* [in] */ long line_no,
			/* [in] */ long mod_no,
			/* [in] */ const std::string &mod_name,
			/* [in] */ const std::string &condition,
			/* [in] */ bool istemp,
			/* [in] */ long skiphits,
			/* [out] */ long *bp_no) = 0;
	virtual HRESULT updatebreakpoint(
			/* [in] */ long bp_no,
			/* [in] */ long line_no,
			/* [in] */ const std::string &condition,
			/* [in] */ long skiphits,
			/* [in] */ BPSTATE state) = 0;
	virtual HRESULT findbreakpoint(
			/* [in] */ long line_no,
			/* [in] */ long mod_no,
			/* [in] */ const std::string &mod_name,
			/* [in] */ const std::string &condition,
			/* [out] */ long *bp_no) = 0;
	virtual HRESULT getbreakpoint(
		/* [in] */ long bp_no,
		/* [out] */ long *line_no,
		/* [out] */ long *mod_no,
		/* [out] */ std::string &mod_name,
		/* [out] */ std::string &condition,
		/* [out] */ bool *istemp,
		/* [out] */ long *skiphits,
		/* [out] */ long *hitcount,
		/* [out] */ BPSTATE *state,
		/* [out] */ bool *isunderhit) = 0;
	virtual HRESULT deletebreakpoint(long bp_no) = 0;
	virtual HRESULT getbreakpointlist(
			/* [in/out] */ BREAKPOINTLIST *bplist) = 0;
/* evaluate */
	virtual HRESULT evaluate(
			/* [in] */ std::string &expr,
			/* [in] */ long scope_id,
			/* [out]*/ std::string *serial_str,
			/* [out]*/ std::string *err_str) = 0;
/* profiler */
	virtual HRESULT getperffreq(
			/* [out] */llong *perffreq,
			/* [out] */long *perfaccuracy) = 0;
	virtual HRESULT getprofdatalist(
            /* [in] */ long mod_no,
			/* [in/out] */ PROFDATALIST *aprofdatalist) = 0;
};


/**************************************************************************

	DbgSiteBase

**************************************************************************/

class DbgSiteBase {
private:
protected:
public:
	virtual ~DbgSiteBase() {}
	virtual HRESULT initialize(RequestorBase *req, int dbg_ver, char *dbg_descr) = 0; 
	virtual HRESULT handleerror(int errtype, char *errmsg, BP_ACTION *action) = 0;  
	virtual HRESULT handlebreakpoint(BR_REASON reason, BP_ACTION *action) = 0;
	virtual HRESULT log(int mod_no, int line_no, char *mod_name, int type, char *logmsg) = 0;
	virtual HRESULT close() = 0;
	virtual HRESULT isalive() {return S_OK;}
	virtual HRESULT delay_handling() = 0;
};


/**************************************************************************

	StringTableItem/StringTable

**************************************************************************/

typedef enum tagMERGETYPE {
	MERGE_OVERWRITE				= 0x0000,
	MERGE_NOOVERWRITE			= 0x0001,
	MERGE_APPEND_W_COMMA		= 0x0002,	// ,
	MERGE_APPEND_W_SEMICOLON	= 0x0003,	// ;
	MERGE_APPEND_W_COLON		= 0x0004,	// :
} MERGETYPE;

#ifdef WIN32
#define MERGE_ENV_VAR MERGE_APPEND_W_SEMICOLON
#else
#define MERGE_ENV_VAR MERGE_APPEND_W_COLON
#endif

#define MERGE_APPEND_MASK		0x000F

extern const char *MERGE_APPEND_CHAR[];

class StringTableItem {
public:
	std::string key;
	std::string val;
	StringTableItem(const char *akey, const char *aval) :
		key(akey), val(aval) {}
	StringTableItem(const std::string& akey, const std::string& aval) :
		key(akey), val(aval) {} 
	StringTableItem() {}
	StringTableItem(const StringTableItem& item) {
		key=item.key;
		val=item.val;
	}
};

typedef std::vector<StringTableItem> StringTableBase;

// ordered array of keys and associated values
class StringTable: protected StringTableBase {
private:
protected:
	void add(const std::string& key, const std::string& val) {
		push_back(StringTableItem(key, val));
	}
	void add(const char *key, const char *val) {
		push_back(StringTableItem(key, val));
	}
	void add(const StringTableItem& item) {
		push_back(item);
	}
public:
	StringTable() {};
	StringTable(const std::string& str, char equ_delimiter = '=', const char *line_delimiter = "\n");
	StringTable(char** strarray, char equ_delimiter = '=');

	iterator begin() { return StringTableBase::begin();}
	const_iterator begin() const {return StringTableBase::begin();}
	iterator end() { return StringTableBase::end();}
	const_iterator end() const {return StringTableBase::end();}
	size_type size() const {return StringTableBase::size();}
	void clear() { StringTableBase::clear();}

	void merge(const std::string& key, const std::string& val, MERGETYPE flags = MERGE_OVERWRITE);
	void merge(const StringTable& atable, MERGETYPE flags = MERGE_OVERWRITE);
	void getline(int index, std::string& str, const char *delimiter="=", const char *line_delimiter="\n") const;
	void getlines(std::string& str, const char *delimiter="=", const char *line_delimiter="\n") const;
};


// simplest array to hold array of pointers, each points to appropriate string <char *arr[]>
class SimpleStringArray {
protected:
	char** farray;
	int size;
public:
	SimpleStringArray() : farray(NULL), size(0) {}
	SimpleStringArray(const SimpleStringArray& src);
	SimpleStringArray(const StringTable& table, char *equ_delimiter = "=");
	~SimpleStringArray() { clear(); }
	void add(char *str);
	void clear();
	int count() const {return size;}
	char * operator [] (int index) { return (index < 0 || index >= size) ? NULL : (farray[index]);}
	const char** array() const {return (const char **) farray;}
	SimpleStringArray& operator = (const SimpleStringArray& src);
};

/**************************************************************************

	SessRequest

**************************************************************************/

class SessRequest {
public:
	SessRequest() : server_type(SVR_AUTO), timeoutms_waitsess(0), timeoutms_client(0), port(0) {}
	SERVER_TYPE server_type;	// CGI or HTTP or HTTPS or AUTO
	int timeoutms_waitsess;		// timeout for session to start
	int timeoutms_client;		// timeout when session works
	unsigned short port;		// listener port
	std::string address;		// local IP address (mostly should be empty!, intended for multi-homed computers only)
	int SOF_flags;
	std::string request;		// url or filename with GET vars
	StringTable headers;		// list of additional request headers
	std::string cgi_handler;	// CGI executable
};


/**************************************************************************

	PipeReader

**************************************************************************/

#ifdef WIN32
#define PIPE_READER_BUF_SIZE 8192
typedef struct tag_read_result {
	int result_code;
	DWORD bytes_read;
	DWORD error_code;
} read_result;

#endif

class PipeReader {
private:
protected:
	pipe_t fpipe;
	int fsocket;
#ifdef WIN32
	HANDLE fthread;
	HANDLE evCanRead,evHasRead,evCanExit;
	struct tag_read_result *pread_result;
	int iopending;
	int bufremaining;
	char *bufhead;
	char *fbuf;
	friend static DWORD WINAPI PipeReader_ThreadProc(void* pv);
#endif
public:
	PipeReader();
	virtual ~PipeReader() {
		close();
	}
	int assignpipe(pipe_t apipe);
	int assignsocket(int asocket);
	void close();

	int recv(char *abuf, int len, int timeoutms);
#ifdef WIN32
	void checkrefill();
#endif

};


/**************************************************************************

	SessConnection

**************************************************************************/

class SessConnection {
public:
	SessConnection() : /*sock(0),*/ result(DBGSE_WRONGREQ), cgi_pid(0) {}
	~SessConnection() {
		if (cgi_pid) {
			char buf[256];
			for (int cnt=0; receiver.recv(buf, sizeof(buf), 100) > 0 && cnt < 100; cnt++) {}
#ifdef WIN32
			HANDLE h = OpenProcess(PROCESS_TERMINATE, FALSE, cgi_pid);
			if (h) {
				TerminateProcess(h, 1);
				CloseHandle(h);
			}
#else
			kill(cgi_pid, SIGTERM);
#endif
		}
	}
	DBGS_RESULT result;
	PipeReader receiver;		// output receiver
	PipeReader errreceiver;		// error stream receiver (valid for CGI only)
	std::string sessid;			// string SESSION ID
	int cgi_pid;
};


/**************************************************************************

	ThreadBase

**************************************************************************/

class ThreadBase;

/* List of clients */
typedef struct thlistitemtag {
	ThreadBase *clientthread;
} thlistitem, *pthlistitem;

typedef std::vector<thlistitem> TH_LIST;

class ThreadBase : public RequestorBase {
private:
protected:
	int fsock;
	int SOF_flags;
	int ftimeoutms_client;
	int flags;
	int siteinitialized;
	int server_flags;
	sockaddr fclient_addr;
	SCOPELIST scopelist;
	BREAKPOINTLIST breakpointlist;
	DOCLIST fdoclist;
	
	DbgSiteBase *fsite;
	SessConnection *fconn;
	char *fconnbuf;
	char *fconnerrbuf;
	int is_valid_bplist;
	
	virtual int ack(int cmd , dbg_packet *pack);
	virtual int ack_action(BP_ACTION action);
	virtual int process_cmd(int cmd, dbg_packet *pack);
	void process_conn();
	virtual DbgSiteBase* create_site() = 0;
	int request(dbg_header_struct *hdr, dbg_packet *pack);
	virtual int onstartsession(dbg_packet *pack);
	virtual int onendsession(dbg_packet *pack);
	int onbreakpoint(dbg_packet *pack, BR_REASON reason);
	int onerror(dbg_packet *pack);
	int onlog(dbg_packet *pack);	
	HRESULT setupdatebreakpoint(
		/* [in] */ long bp_no_in,
		/* [in] */ long line_no,
		/* [in] */ long mod_no,
		/* [in] */ const std::string &mod_name,
		/* [in] */ const std::string &condition,
		/* [in] */ bool istemp,
		/* [in] */ long skiphits,
		/* [in] */ BPSTATE state,
		/* [out] */ long *bp_no_out);

public:
	ThreadBase(): fsock(0), SOF_flags(0),
		ftimeoutms_client(0), flags(0),
		siteinitialized(0), server_flags(0),
		fsite(0), fconn(0), fconnbuf(0), fconnerrbuf(0),
		is_valid_bplist(0) {
		DBGTRACE(("ThreadBase::ThreadBase\n"));
		memset(&fclient_addr, 0, sizeof(fclient_addr));
	}
	virtual ~ThreadBase() {}
	// JIT
	virtual bool assign_connection(
		int sock,
		sockaddr *client_addr,
		int aSOF_flags,
		int timeoutms_client,
		DbgSiteBase *asite = NULL,
		SessConnection *aconn = NULL);
	// REQ
	virtual void close();
	virtual void process();
	char isfinished();
	virtual void stop(bool bWait) {} // shoud be realized with MultiThreaded listener
	virtual void destroy() {
		stop(true);
		delete this; // should be replaced with _Release() in COM case.
	}
	int sock() {return fsock;}

//////////////// RequestorBase PART ////////////////////
/* info */
	HRESULT getdbgversion(
			/* [out] */ long *dbg_ver,
			/* [out] */ std::string *sdbg_descr);
	HRESULT getserveraddr(
			/* [out] */ sockaddr_in *addr);
	HRESULT setsessionoptions(
			/* [in] */ long opt_flags);
/* modules */
	HRESULT getdoclist(
			/* [in/out] */ DOCLIST *doclist);
    HRESULT requestdocsrc(
            /* [in] */  long mod_no,
            /* [out] */ std::string *mod_name,
            /* [out] */ std::string *src);
    HRESULT requestdocparent( 
            /* [in] */ long mod_no,
            /* [out] */ long *parent_mod_no);
    HRESULT requestdocfilename( 
            /* [in] */ long mod_no,
            /* [out] */ std::string *mod_name);
	HRESULT getlinesinfolist(
            /* [in] */ long mod_no,
			/* [in/out] */ LINESINFOLIST *linesinfolist);
	HRESULT getctxinfolist(
            /* [in] */ long mod_no,
			/* [in/out] */ CTXINFOLIST *ctxinfolist);

/* scopes */
	HRESULT getscopelist(
			/* [in] */ SCOPELIST *ascopelist);
	HRESULT getcurloc(
			/* [out] */ long *mod_no,
			/* [out] */ long *line_no,
			/* [out] */ std::string *scope_name);
/* breakpoints */
	HRESULT setbreakpoint(
			/* [in] */ long line_no,
			/* [in] */ long mod_no,
			/* [in] */ const std::string &mod_name,
			/* [in] */ const std::string &condition,
			/* [in] */ bool istemp,
			/* [in] */ long skiphits,
			/* [out] */ long *bp_no);
	HRESULT updatebreakpoint(
			/* [in] */ long bp_no,
			/* [in] */ long line_no,
			/* [in] */ const std::string &condition,
			/* [in] */ long skiphits,
			/* [in] */ BPSTATE state);
	HRESULT findbreakpoint(
			/* [in] */ long line_no,
			/* [in] */ long mod_no,
			/* [in] */ const std::string &mod_name,
			/* [in] */ const std::string &condition,
			/* [out] */ long *bp_no);
	HRESULT getbreakpoint(
			/* [in] */ long bp_no,
			/* [out] */ long *line_no,
			/* [out] */ long *mod_no,
			/* [out] */ std::string &mod_name,
			/* [out] */ std::string &condition,
			/* [out] */ bool *istemp,
			/* [out] */ long *skiphits,
			/* [out] */ long *hitcount,
			/* [out] */ BPSTATE *state,
			/* [out] */ bool *isunderhit);
	HRESULT deletebreakpoint(
			/* [in] */ long bp_no);
	HRESULT getbreakpointlist(
			/* [in/out] */ BREAKPOINTLIST *bplist);
/* evaluate */
	HRESULT evaluate(
			/* [in] */ std::string &expr,
			/* [in] */ long scope_id,
			/* [out]*/ std::string *serial_str,
			/* [out]*/ std::string *err_str);
/* profiler */
	HRESULT getperffreq(
			/* [out] */llong *perffreq,
			/* [out] */long *perfaccuracy);
	HRESULT getprofdatalist(
            /* [in] */ long mod_no,
			/* [in/out] */ PROFDATALIST *aprofdatalist);
};



/**************************************************************************

	ListenerBase

**************************************************************************/

class ListenerBase {
private:
protected:
	int fsocketJIT;
	cfgprm fcfgJIT;
	ListenerStatus fstatusJIT;
	int ferrorJIT;
	TH_LIST th_list;
	virtual ThreadBase *create_client() = 0; // pure method
	virtual void destroy_client(ThreadBase *client) {
		client->destroy();
	}

public:
	ListenerBase(): fsocketJIT(0), fstatusJIT(LSTNRS_notinitialized), ferrorJIT(0) {
		DBGTRACE(("ListenerBase::ListenerBase\n"));
		//memset(&fcfgJIT, 0, sizeof(fcfgJIT));
	}
	virtual ~ListenerBase();

// Clients
	int numberofclients() {	return th_list.size();}
	ThreadBase *client(int index) {
		return (index >= 0 && index < (int)th_list.size()) ? th_list[index].clientthread : NULL;
	}
	virtual void stop(bool stopclients);

// JIT debugging support
	virtual int startJIT(cfgprm *cfg, void *extcfg);
	virtual int restartJIT(cfgprm *newcfg, void *extcfg);
	virtual void stopJIT();
	virtual int peekJIT();
	virtual void statuschangedJIT() = 0;	
	ListenerStatus statusJIT() { return fstatusJIT; }
	int errorJIT() { return ferrorJIT; }
	void logstatusJIT();

// REQ debugging support
	virtual HRESULT startsession(SessRequest& sessionreq, DbgSiteBase *fsite, DBGS_RESULT *result);

// misc logging service
	virtual void log(ListenerLogKind kind, const char *message,...) = 0;
	virtual void logsyserror(const char* message, int err=0);
};


#endif
