/***************************************************************************
                          dbgbase_globals.h  -  description
                             -------------------
    begin                : Fri Mar 2 2001
    copyright            : (C) 2001 by Dmitri Dmitrienko
    email                : dd@cron.ru
    www                  : http://dd.cron.ru
    license              : This source file is subject to version 1.01 of 
                           the License,  that is bundled with this package 
                           in the file LICENSE, and is available at through 
                           the world-wide-web at http://dd.cron.ru/license
 ***************************************************************************/

#ifndef _DBGBASE_GLOBALS_H_
#define _DBGBASE_GLOBALS_H_


#ifdef WIN32
#include <Windows.h>
#else
#include <stdio.h>
#include <sys/param.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include <vector>
#include <string>

#include <dbgbase/dbg_net.h>
#include <dbgbase/dbgbase_intf.h>


#ifdef WIN32
  #include <Winerror.h>
#else
  #ifndef _HRESULT_DEFINED
    #define _HRESULT_DEFINED
    typedef long HRESULT;
	#define S_OK							((HRESULT)0x00000000L)
	#define S_FALSE							((HRESULT)0x00000001L)
	#define E_FAIL							((HRESULT)0x80004005L)
	#define E_POINTER						((HRESULT)0x80004003L)
	#define SUCCEEDED(Status)				((HRESULT)(Status) >= 0)
	#define FAILED(Status)					((HRESULT)(Status) < 0)
  #endif
#endif

#ifdef WIN32
/* Win 32 */
typedef __int64 llong;
#else
/* Linux / Unix */
#include <time.h>
typedef long long llong;
#endif


#ifdef WIN32
  #define	snprintf _snprintf
  #define	vsnprintf _vsnprintf
#else
  #define	stricmp strcasecmp
#endif

#ifndef SOCKET_ERROR
#define SOCKET_ERROR (-1)
#endif

#ifndef SD_BOTH
#define SD_BOTH SHUT_RDWR
#endif

#ifdef _TRACE
  typedef void (*logger_cb_t)(char *buf);
  extern logger_cb_t tracelogger;

  void __dbgtrace(const char *str,...);
  #define DBGTRACE(p) __dbgtrace p
#else
  #define DBGTRACE(p)
#endif

typedef struct cfgprm_tag {
	cfgprm_tag() : timeoutms_client(0), timeoutms_process(0), port(0), breakonload(0), breakonfinish(0) {}
	int timeoutms_client;
	int timeoutms_process;
	unsigned short port;
	int breakonload;
	int breakonfinish;
	std::string address;
} cfgprm;

#define DEFAULT_LSTNR_PORT 7869
#define DEFAULT_LSTNR_TIMEOUT 30000 /* 30 sec */
#define DEFAULT_LSTNR_ADDRESS ""
#define DEFAULT_LSTNR_BREAKONSTART 1
#define DEFAULT_LSTNR_BREAKONFINISH 1

#define DEFAULT_REQ_PORT	10001
#define REQ_PORT_COUNT		100
#define DBGSESSVAR			"DBGSESSID"


#define CONNBUFSIZE 8192
#define CONNERRBUFSIZE 8192

////////////////// P I P E S ///////////////////////////////////

#ifdef WIN32

typedef HANDLE			pipe_t;

#define INVALID_PIPE	INVALID_HANDLE_VALUE

#define CLOSEPIPE(a)		\
	if (a!=INVALID_PIPE) {	\
		CloseHandle(a);		\
		a = INVALID_PIPE;	\
	}

#define CLOSEHANDLE(a)		\
	if ((a)!=NULL) {		\
		CloseHandle(a);		\
		a = NULL;			\
	}


#else

typedef int				pipe_t;

#define INVALID_PIPE	-1

#define CLOSEPIPE(a)		\
	if (a!=INVALID_PIPE) {	\
		::close(a);			\
		a = INVALID_PIPE;	\
	}

#endif

////////////////// H E L P E R   C L A S S E S ///////////////////////////////////

/* List of active scopes */
typedef struct scopelistitemtag {
	long scope_id;
	long mod_no;
	long line_no;
	std::string scope_descr;
} scopelistitem, *pscopelistitem;

typedef std::vector<scopelistitem> SCOPELISTBASE;

class SCOPELIST: public SCOPELISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const SCOPELIST &scopelist) {
		clear();
		insert(begin(), scopelist.begin(), scopelist.end());
	}
};

/* List of breakpoints */
typedef struct breakpointlistitemtag {
	int bp_no;
	long line_no;
	long mod_no;
	std::string mod_name;
	std::string condition;
	bool istemp;
	int skiphits;
	int hitcount;
	BPSTATE state;
	bool isunderhit;
} breakpointlistitem, *pbreakpointlistitem;

typedef std::vector<breakpointlistitem> BREAKPOINTLISTBASE;

class BREAKPOINTLIST: public BREAKPOINTLISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const BREAKPOINTLIST &list) {
		clear();
		insert(begin(), list.begin(), list.end());
	}
	void add(
		long bp_no,
		long line_no,
		long mod_no,
		char *mod_name,
		char *condition,
		bool istemp,
		int skiphits,
		int hitcount,
		BPSTATE state,
		bool isunderhit);
	void update(
		long bp_no,
		long line_no,
		long mod_no,
		char *mod_name,
		char *condition,
		bool istemp,
		int skiphits,
		int hitcount,
		BPSTATE state,
		bool isunderhit);
};

/* List of documents */
typedef struct doclistitemtag {
	long mod_no;
	long parent_mod_no;
	std::string mod_name;
} doclistitem, *pdoclistitem;

typedef std::vector<doclistitem> DOCLISTBASE;

class DOCLIST: public DOCLISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const DOCLIST &list) {
		clear();
		insert(begin(), list.begin(), list.end());
	}
};


/* List of linesinfo */
typedef struct linesinfoitemtag {
	long mod_no;
	long start_line_no;
	long lines_count;
	long ctx_id;
} linesinfoitem, *plinesinfoitem;

typedef std::vector<linesinfoitem> LINESINFOLISTBASE;

class LINESINFOLIST: public LINESINFOLISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const LINESINFOLIST &li) {
		clear();
		insert(begin(), li.begin(), li.end());
	}
};


/* List of ctxinfo */
typedef struct ctxinfoitemtag {
	long mod_no;
	std::string function_name;
	long ctx_id;
} ctxinfoitem, *pctxinfoitem;

typedef std::vector<ctxinfoitem> CTXINFOLISTBASE;

class CTXINFOLIST: public CTXINFOLISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const CTXINFOLIST &li) {
		clear();
		insert(begin(), li.begin(), li.end());
	}
};


/* List of profdata */
typedef struct profdataitemtag {
	long mod_no;
	long line_no;
	long hit_count;
	llong tm_min;
	llong tm_max;
	llong tm_sum;
} profdataitem, *pprofdataitem;

typedef std::vector<profdataitem> PROFDATALISTBASE;

class PROFDATALIST: public PROFDATALISTBASE {
public:
	void assign(dbg_packet *pack);
	void assign(const PROFDATALIST &li) {
		clear();
		insert(begin(), li.begin(), li.end());
	}
};

/* List of eval data */
typedef enum php_typetag {
	ptUnknown, 
	ptLong, 
	ptDouble, 
	ptString,
    ptArray, 
	ptObject, 
	ptBool, 
	ptResource, 
	ptRef, 
	ptSoftRef
} php_type;

class EVALLIST;

extern char *valtypes[10];

typedef struct eval_itemtag {
	std::string name;
	std::string value;
	std::string valtype;
	php_type valtypeid;
	EVALLIST* subitems;
	int var_index;
	struct eval_itemtag *ref;
	eval_itemtag() : valtypeid(ptUnknown), subitems(NULL), var_index(0), ref(NULL) {}
	eval_itemtag(const std::string& aname, const std::string& avalue, const std::string& avaltype,
		php_type avaltypeid, EVALLIST* asubitems) : 

		name(aname), value(avalue), valtypeid(avaltypeid), subitems(asubitems),
		var_index(0), ref(NULL)
		
	{
		if (avaltype.length() > 0) {
			valtype=avaltype;
		} else {
			valtype=valtypes[avaltypeid];
		}
	}
	~eval_itemtag();
} eval_item, *peval_item;

typedef std::vector<peval_item> EVALLISTBASE;
typedef std::vector<peval_item> VARLIST;


class EVALLIST: public EVALLISTBASE {
public:	
	EVALLIST(const std::string& serial_str, const std::string& name);
	EVALLIST() {};
	eval_item *Add(VARLIST* var_list, const std::string& name, const std::string& value, 
		php_type valtypeid = ptUnknown, const std::string& valtype = "", EVALLIST* subitems=NULL) {

		peval_item it;

		it = new eval_item(name, value, valtype, valtypeid, subitems);
		push_back(it);
		if (var_list) var_list->push_back(it);
		return it;
	}
	~EVALLIST() {
		clear();
	};
	void clear() {
		for (const_iterator it=begin(); it!=end(); it++) {
			delete *it;
		}
		EVALLISTBASE::clear();
	}
};

// FILEMAPPER


typedef enum tag_map_location {
	mlLocal, 
	mlRemote, 
	mlURL,
} map_location;


typedef enum tag_map_kind {
    mkLocalToRemote, 
	mkLocalToURL,
    mkRemoteToLocal, 
	mkRemoteToURL,
    mkURLToLocal,    
	mkURLToRemote,
} map_kind;


#ifdef WIN32
#define LOCAL_FS_SLASH '\\'
#define REMOTE_FS_SLASH '\\'
#else
#define LOCAL_FS_SLASH '/'
#define REMOTE_FS_SLASH '/'
#endif


class filemapper {
private:
	std::string furl_root;
	std::string flocal_root;
	std::string fremote_root;
public:
	bool is_HTTP;
	filemapper(bool mode_is_HTTP, std::string aurl_root, std::string alocal_root, std::string aremote_root) : 
		furl_root(aurl_root), flocal_root(alocal_root), fremote_root(aremote_root), is_HTTP(false) {}
	bool mapfilename(map_kind type, std::string from, std::string& to, std::string& warnstr);
	void modified();
	void set_url_root(std::string s) {
		furl_root = s;
		modified();
	}
	void get_url_root(std::string& s) {
		s = furl_root;
	}
	void set_local_root(std::string s) {
		flocal_root = s;
		modified();
	}
	void get_local_root(std::string& s) {
		s = flocal_root;
	}
	void set_remote_root(std::string s) {
		fremote_root = s;
		modified();
	}
	void get_remote_root(std::string& s) {
		s = fremote_root;
	}
};

// Misc functions

int sprintf(std::string& buf, const char *format ...);
void dbgtoupper(const std::string& str, std::string& buf);
bool fileexists(std::string& fname);
bool findfile(std::string& fname);

#ifdef WIN32
int gettimeofday(struct timeval *ti);
#define GETTIMEOFDAY(ti) gettimeofday(ti)
char *realpath(const char *fname, char *realname);
#else
#define GETTIMEOFDAY(ti) gettimeofday(ti, NULL)
#endif

bool realpath(std::string &fname, std::string &realname);

/////////////////////////// O S - R E L A T E D ///////////////////////////
/*             OS-SPECIFIC ERROR             */
int oserrno(int isnetwork);
void oserrnostr(std::string& buf, int err=-1);


/*             LISTENER SOCKET               */
int dbh_create_listen_socket(const char* addr, unsigned short port );
int dbh_check_new_client_connect(int listen_socket, sockaddr *addr, int timeoutms);
int dbh_wait_debug_event(dbg_header_struct *hdr, int client_sock, dbg_packet *pack, int timeoutsec);
void dbh_close_socket(int socket);
int dbh_ack(int cmd , dbg_packet *pack, int socket, int flags);

#endif
